#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Connections

Represents a generic gateway to AMQP implementation
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.

__all__ = ['ConnectionFactory']

# Project requirements
from sleipnir.core.factory import AbstractFactory
from sleipnir.core.decorators import deprecated

# local submodule requirements
from .enums import ConnectionType, ConnectionState
from .parser import connection_url, URI


class ConnectionError(Exception):
    """Connection exception"""


class ConnectionFactory(AbstractFactory):
    """Connection factory"""

    _shared_backends = {}
    _shared_connections = {}

    def __getattribute__(self, name):
        if name == '_backends':
            name = '_shared_backends'
        if name == '_connections':
            name = '_shared_connections'
        return object.__getattribute__(self, name)

    def __getitem__(self, key):
        return self._connections.get(key, None)

    def __delitem__(self, key):
        if key in self._connections:
            self._connections[key].close()
        del self._connections[key]

    def _on_connected(self, conn):
        conn.state = ConnectionState.STARTED
        if conn.url in self._connections:
            # invoke template method
            self.when_connected(conn)
            # add callbacks for close and backpressure events
            conn.add_on_close_callback(self._on_disconnected)
            conn.add_backpressure_callback(self._on_backpressure)

    def _on_disconnected(self, conn):
        conn.state = ConnectionState.CLOSED
        if conn.url in self._connections:
            self.when_disconnected(conn)

    def _on_backpressure(self, conn):
        if conn.url in self._connections:
            if not conn.push_timeout:
                conn.push_timeout = BACKPRESSURE
            conn.push_timeout *= BACKPRESSURE_MULTIPLIER
            conn.push_timeout = min(conn.push_timeout, BACKPRESSURE_LIMIT)
            # notify about pressure
            self.when_backpressure(conn)

    def _on_connection_timeout(self):
        for conn in self._connections.values():
            if conn.state == ConnectionState.STARTING:
                self.when_error(conn)
                del self._connections[conn.url]

    def when_connected(self, connection):
        pass

    def when_disconnected(self, connection):
        pass

    def when_backpressure(self, connection):
        pass

    def when_error(self, connection):
        pass

    def create(self, url, conn_type=ConnectionType.AUTO_ASYNC):
        # check that url pased (url_string, connection or channel)
        # doesn't exist in connections pool
        connection = url
        if hasattr(connection, 'slp_connection'):
            connection = url.slp_connection
        if isinstance(connection, AbstractConnection):
            url = connection.url

        # handle auto case
        if conn_type == ConnectionType.AUTO_ASYNC:
            if url.startswith("amqp"):
                conn_type = ConnectionType.AMQP_ASYNC
            elif url.startswith("http"):
                conn_type = ConnectionType.HTTP_ASYNC
            else:
                raise NotImplementedError
            
        # check now url
        if url not in self._connections:
            conn = super(ConnectionFactory, self).create(url, conn_type)
            self._connections[url] = conn
            # set callbacks
            conn.add_on_open_callback(self._on_connected)
            # add a timeout so if connection isn't valid, we should remove it
            conn.add_timeout(2, self._on_connection_timeout)
            conn.state = ConnectionState.STARTING
        return self._connections[url]


class MetaConnection(type):
    """Connection Metaclass"""

    def __init__(mcs, name, bases, dct):
        type.__init__(mcs, name, bases, dct)
        if not name.endswith("Connection"):
            ConnectionFactory.get_instance().register(name, mcs)


class AbstractConnection(object):
    """Prepares a connection. only get a pika connection is lazy created"""

    __metaclass__ = MetaConnection

    def __init__(self, url, limit=4):
        """Whenever a property changes, update stamp"""
        self._url = URI(url, connection_url)
        self._state = ConnectionState.CLOSED

    @property
    def type(self):
        return self.CONNECTION_TYPE

    @property
    def url(self):
        return str(self._url)

    @property
    def loop(self):
        raise NotImplementedError

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    @classmethod
    def can_handle(cls, url, conn_type):
        return cls.CONNECTION_TYPE == conn_type

    @classmethod
    def new(cls, url, conn_type):
        return cls(url)

# load connection implementations
from .amqp.connections import *
from .http.connections import *



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Connections

Represents a generic gateway to AMQP implementation
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
from pika import PlainCredentials, ConnectionParameters
from pika.adapters import SelectConnection

__all__ = []

# Project requirements

# local submodule requirements
from ..enums import ConnectionType
from ..connections import AbstractConnection
from .pools import ChannelPool, BuilderPool

BACKPRESSURE = 0.15
BACKPRESSURE_LIMIT = 90
BACKPRESSURE_MULTIPLIER = 3


class AMQPConnection(AbstractConnection):
    """Prepares a connection. only get a pika connection is lazy created"""

    def __init__(self, url, limit=4):
        """Whenever a property changes, update stamp"""
        AbstractConnection.__init__(self, url)
        # AMQP backpressure
        self._push_timeout = 0
        # set up pools
        self._builders_pool = BuilderPool(self)
        self._channels_pool = ChannelPool(self, limit)

    @property
    def channels(self):
        return self._channels_pool

    @property
    def builders(self):
        return self._builders_pool

    @property
    def loop(self):
        # get connection variable to be used by start method
        connection = self

        def start(self, timeout=None):
            if timeout is not None:
                connection.add_timeout(timeout, connection.close)
            return self.__class__.start(self)

        # Start event loop
        if not hasattr(self.ioloop, "slp_monkey_patched"):
            self.ioloop.slp_monkey_patched = True
            klass = self.ioloop.__class__
            func_type = type(klass.start)
            self.ioloop.start = func_type(start, self.ioloop, klass)
        return self.ioloop

    @property
    def push_timeout(self):
        return self._push_timeout

    @push_timeout.setter
    def push_timeout(self, value):
        self._push_timeout = value

    def routes(self, builder, **builder_kwargs):
        builder = builder(self, **builder_kwargs)
        return self._builders_pool.register(builder)


class ConnectionMockAMQP(AMQPConnection):
    """A mock connection object"""

    CONNECTION_TYPE = ConnectionType.AMQP_MOCK

    class IOLoopMock(object):
        def start(self):
            pass

    @property
    def loop(self):
        return IOLoopMock()


class PikaConnection(AMQPConnection):

    def __init__(self, url):
        AMQPConnection.__init__(self, url)

        credentials = None
        if self._url.username:
            credentials = PlainCredentials(
                self._url.username, self._url.password)

        for host, port in self._url.hosts():
            self._parameters = ConnectionParameters(
                credentials=credentials,
                host=host,
                port=port,
                virtual_host=self._url.vhost)

            #Fixme: Support multiple hosts
            break


class ConnectionAMQP(PikaConnection, SelectConnection):

    CONNECTION_TYPE = ConnectionType.AMQP_ASYNC

    def __init__(self, url):
        PikaConnection.__init__(self, url)
        SelectConnection.__init__(self, self._parameters)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
exceptions

Common exceptions used by transport modules
"""

from __future__ import absolute_import

__author__  = "Carlos Martin"
__license__ = "BSD, See LICENSE file for details"

# Import here any required modules.
import sleipnir.core

__all__ = [
    'ConnectionError',
    ]


class ConnectionError(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __str__(self):
        return "{0} {1}".format(self.code, self.message)


class MessageStateError(Exception):
    """The message has already been acknowledged."""
    pass



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Aggregator

Base classes to implement message endpoints (EIP2003)
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.

__all__ = ['Aggregator']

# Project requirements
from sleipnir.core.utils import maybe_list

# local submodule requirements
from ..connections import ConnectionFactory
from .endpoints import EndPoint


class Comparator(object):

    def __init__(self, comparator):
        self._comparator = comparator
        self._operator = None
        self._value = None

    def __call__(self, body, message):
        raise NotImplementedError

    def compare(self, operator, value):
        self._operator = operator
        self._value = value
        return self

    @property
    def operator(self):
        return self._operator

    @property
    def value(self):
        self._value


class JPath(Comparator):

    import jsonpath

    def __init__(self, path):
        super(JPath, self).__init__(path)
        self._comparator = jsonpath.normalize(path)

    def __call__(self, body, message):
        match, result = None, jsonpath.jsonpath(dictionary, self._path)
        if callable(self._operator):
            match = self._operator(result, self._value)
        return result, match


class Constant(Comparator):
    def __init__(self, value=True):
        super(Constant, self).__init__(value)

    def __call__(self, body, message):
        return self._comparator, None


class Header(Comparator):
    def __call__(self, body, message):
        return (message.properties.get(self._comparator, None), None,)


# merge agregate and agregate strategy ?
class Aggregate(object):

    __slots__ = ('_request', '_message', '__weakref__')

    def __init__(self):
        self._request = None
        self._message = None

    def add_message(self, request, message):
        raise NotImplementedError

    def get_message(self):
        return self._request, self._message


class WrapperAggregate(Aggregate):

    __slots__ = ('_wrapper', '_request')

    # Wrappable Class
    wrapper = None

    def __init__(self, *args):
        super(WrapperAggregate, self).__init__()
        self._wrapper = self.wrapper(*args)
        self._message = None

    def add_message(self, request, message):
        self._message = self._message or message
        self._wrapper.add_message(message.payload)

    def is_complete(self, *args):
        return self._wrapper.is_complete(*args)

    def get_message(self):
        return self._wrapper.get_message(), self._message


# class AuctionAggregate(WrapperAggregate):

#     __slots__ = ()

#     # wrapped class
#     wrapper = Bid


# class Auction(object):

#     def __init__(self):
#         self.__bids = []

#     def add_message(self, message):
#         self.__bids.append(message)

#     def is_complete(self):
#         return len(self.__bids) > 3

#     def get_message(self):
#         winner = None
#         for bid in self.__bids:
#             winner = winner or bid
#             if int(bid['PRICE']) < int(winner['PRICE']):
#                 winner = bid
#         return winner


class LastOneAggregate(Aggregate):
    def add_message(self, request, message):
        self._request, self._message = request, message


class AggregatorFlags(enum):
    TIMEOUT,   \
    INTERVAL,  \
    SIZE,      \
    PREDICATE, \
    AGGREGATE = xrange(0, 5)


class CompletionAbstract(object):

    flag = None
    partial = None

    def __init__(self, container, value=None):
        self._container = container
        value = [self._check, self.partial, value or []]
        self._container._flags[self.flag] = value

    def _check(self, agregate):
        raise NotImplementedError


class CompletionTimeout(CompletionAbstract):

    flag = AggregatorFlags.TIMEOUT
    partial = False

    def __init__(self, container, value):
        super(CompletionTimeout, self).__init__(container, value)
        self._container.connection.add_timeout(value, self.__on_timeout)

    def __on_timeout(self):
        self._container.flags[self.flag][2] = True
        self._container.check()

    def _check(self, aggregate):
        return self._container.flags[self.flag][2] is True


class CompletionPredicate(CompletionAbstract):

    flag = AggregatorFlags.PREDICATE
    partial = True

    def _check(self, aggregate):
        args = self._container._flags[self.flag][2]
        return aggregate.is_complete(*args)


class AbstractAggregator(object):

    aggregate = LastOneAggregate

    def __init__(self, correlation, aggregate=None):

        # Aggregates storage
        self._flags = {}
        self._aggregates = {}
        self._reverse_aggregates = {}

        # Correlation stuff
        self._correlation = correlation
        self._aggregate = aggregate or self.aggregate

        # completion stuff
        self._predicate = None

        # check if whe should call is_complete for every aggregate
        # defined. This is a defualt handler. If you want to use args
        # or kwargs on is_complete method, use a CompletePredicate
        # decorator wich will override this entry
        if hasattr(self._aggregate, "is_complete"):
            CompletionPredicate(self)

    def __iter__(self):
        return self._aggregates.iteritems()

    def __len__(self):
        return len(self._aggregates)

    def _reply_on_completion(self, aggregate=None):
        # @partial: whether a match should clear all agregates and
        # start again. A false value indicates that all agregates
        # should be removed on a match
        for check, partial, value, in self._flags.itervalues():
            if check(aggregate):
                if partial:
                    aggregate_key = self._reverse_aggregates[aggregate]
                    del self._reverse_aggregates[aggregate]
                    del self._aggregates[aggregate_key]
                    self.reply(*aggregate.get_message())
                else:
                    aggrs = self._aggregates.itervalues()
                    [self.reply(*aggr.get_message()) for aggr in aggrs]
                    self.clear()
                break

    def create(self, aggregate_key, *args, **kwargs):
        aggregate = self._aggregates.get(aggregate_key)
        if not aggregate:
            aggregate = self._aggregate(*args, **kwargs)
            self._aggregates[aggregate_key] = aggregate
            self._reverse_aggregates[aggregate] = aggregate_key
        return aggregate

    def clear(self):
        self._aggregates.clear()
        self._reverse_aggregates.clear()

    def check(self, aggregate=None):
        # check completion
        aggregate = aggregate or self._aggregates.itervalues()
        [self._reply_on_completion(aggr) for aggr in maybe_list(aggregate)]

    def process(self, request, message=None):
        # get correlation key
        corkey, _ = self._correlation(request, message)
        # get aggregate, parse and check completion
        aggregate = self.create(corkey)
        aggregate.add_message(request, message)
        self._reply_on_completion(aggregate)

    def reply(self, request, message=None):
        raise NotImplementedError


class Aggregator(AbstractAggregator):
    def __init__(self, correlation, callback, aggregate=None):
        super(Aggregator, self).__init__(correlation, aggregate)
        self._callback = callback

    def reply(self, response, message):
        callbacks = maybe_list(self._callback)
        [callback(response, message) for callback in callbacks]



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Connections

Represents a connection to an HTTP resource
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.

__all__ = []

# Project requirements

# local submodule requirements
from ..connections import AbstractConnection


class HTTPConnection(AbstractConnection):
    """Prepares a connection """

    def __init__(self, url, template=None):
        AbstractConnection.__init__(self, url)

    @property
    def loop(self):
        raise NotImplementedError



#!/usr/bin/env python
# -*- mode: Python; indent-tabs-mode: nil; coding: utf-8  -*-

#
# Copyright 2011, 2013 Carlos Martín
# Copyright 2011, 2013 Universidad de Salamanca
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

"""Main Setup File"""

import os
import re
import sys

if sys.version_info < (2, 6):
    sys.stderr.write("Sleipnir requires Python 2.6 or higher.")
    sys.exit(0)

# Fix namespace for partial installations
sys.path.insert(0, (os.path.dirname(__file__) or os.curdir) + os.sep + 'src')
from pkg_resources import fixup_namespace_packages
fixup_namespace_packages(sys.path[0])

# setuptools build required.
from setuptools import setup, find_packages

# Load constants
from sleipnir.transport import constants

# Peek author details
AUTHOR, EMAIL = re.match(r'^(.*) <(.*)>$', constants.__author__).groups()

setup(
    author             = AUTHOR,
    author_email       = EMAIL,
    classifiers        = constants.__classifiers__,
    description        = constants.__summary__,
    install_requires   = constants.__requires__,
    license            = constants.__license__,
    long_description   = constants.__long_description__,
    name               = constants.__appname__,
    namespace_packages = [constants.__namespace__],
    package_dir        = {'': 'src'},
    packages           = find_packages(where=sys.path[0], exclude=['*.tests']),
    test_suite         = constants.__appname__ + '.tests',
    tests_require      = [constants.__tests_requires__],
    url                = constants.__url__,
    version            = constants.__version__,
    zip_safe           = False,
)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Translators

Bank instance that process loan propositions
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.

__all__ = ['Translator']

# Project requirements
from sleipnir.core.decorators import cached

# local submodule requirements
from itertools import izip_longest, imap, ifilter


class Translator(object):
    def __init__(self, klass=None):
        self._klasses = {}
        self._current = klass

    def __call__(self, klass):
        self._current = klass
        return self

    @cached
    def __convert(self, klass, *values):
        def fetch_class(klass):
            try:
                klass = eval(klass)
                return klass
            except:
                import inspect
                stack = inspect.stack()
                for level in xrange(len(stack) - 1, 0, -1):
                    try:
                        return stack[level][0].f_locals[klass]
                    except:
                        continue
            return None

        # init result file
        result = dict(izip_longest(imap(str, iter(klass)), (), fillvalue=None))
        # try automated adapt
        for value in ifilter(lambda x: len(x), values):
            for element, element_value in value.iteritems():
                try:
                    key = element.split('.')[1] if '.' in element else element
                    result[str(self._current(key))] = element_value
                except:
                    continue
        # use template method if any
        template = self._klasses.get(klass)
        if callable(template):
            result = template(result, klass, *values)
        return result

    def register(self, klass, template=None):
        self._klasses.setdefault(klass, template)

    def convert(self, *values):
        if self._current:
            return self.__convert(self._current, *values)
        raise NotImplementedError



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Pika

A set of custom classes to better integration of pika with PySide
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
from itertools import ifilter

__all__ = ['PySideReconnectionStrategy', 'PySideConnection']

# Pyside requirements
from PySide.QtCore import QObject, QSocketNotifier, QCoreApplication, QTimer
from PySide.QtNetwork import QNetworkConfigurationManager, QNetworkSession

# Project requirements
from sleipnir.core.decorators import cached

#Pika requirements
from pika.reconnection_strategies import ReconnectionStrategy
from pika.adapters.base_connection import BaseConnection, READ, WRITE


class PySideReconnectionStrategy(ReconnectionStrategy):

    can_reconnect = True

    def __init__(self):
        self.manager = QNetworkConfigurationManager()
        self.session = self._session()
        self.manager.onlineStateChanged.connect(self._connect)

    def _session(self):
        return QNetworkSession(self.manager.defaultConfiguration())

    def _connect(self, is_connected):
        if is_connected is False:
            # create a new session
            self.session = self._session()
            # start session if required
            caps = self.manager.capabilities()
            if caps & QNetworkConfigurationManager.CanStartAndStopInterfaces:
                self.session.open()
                self.session.waitForOpened(-1)

    def on_connect_attempt(self, conn):
        self._connect(self.manager.isOnline())

    def on_connection_open(self, conn):
        caps = self.manager.capabilities()
        if caps & QNetworkConfigurationManager.ForcedRoaming:
            reconnect = conn.force_reconnect
            self.session.newConfigurationActivated.connect(reconnect)

    def on_connection_closed(self, conn):
        conn._reconnect() if not self.is_active else None


class PySideTimer(object):
    def __init__(self, container, callback, single_shot):
        self.callback = callback
        self.single_shot = single_shot
        self.first_run = False

    def register(self, pool, deadline):
        timeout_id = pool.startTimer(deadline)
        pool.timers[timeout_id] = self
        return timeout_id

    def unregister(self, pool, timeout_id):
        pool.killTimer(timeout_id)
        del pool[timeout_id]

    def __call__(self, pool, timeout_id):
        self.callback()
        if self.single_shot:
            self.unregister(pool, timeout_id)


class PySideConnectionPoller(QObject):

    def __init__(self, connection):
        # Set container
        self.parent = connection

    def __iter__(self):
        return iter((self.reader, self.writer,))

    def _connect(self, notifier_type, callback):
        notifier = QSocketNotifier(self.parent.fileno, notifier_type)
        notifier.activated.connect(callback)
        notifier.setEnabled(False)
        return notifier

    def _read(self, _):
        self.parent._handle_read()
        self.parent._manage_event_state()

    def _write(self, _):
        self.parent._handle_write()
        self.parent._manage_event_state()

    def _error(self, _):
        self.parent._handle_disconnect()

    def poll(self):
        # Create Notifiers
        self.reader = self._connect(QSocketNotifier.Read,  self._read)
        self.writer = self._connect(QSocketNotifier.Write, self._write)
        # Create Error watcher
        self.errors = self._connect(QSocketNotifier.Exception, self._error)
        self.errors.setEnabled(True)
        # update handlers
        self.parent.ioloop.update_handler(None, self.parent.event_state)

    def unpoll(self):
        self.reader = self.writer = self.errors = None


class PySideConnection(BaseConnection):

    def __iter__(self):
        return iter(self.notifiers)

    def _adapter_connect(self):
        # Connect (blockignly!) to the server
        BaseConnection._adapter_connect(self)
        self.event_state |= WRITE
        # Setup the IOLoop
        self.ioloop = IOLoop(self.notifiers)
        # Let everyone know we're connected
        self._on_connected()

    def _flush_outbound(self):
        self._manage_event_state()

    @property
    def fileno(self):
        return self.socket.fileno()

    @property
    @cached
    def notifiers(self):
        return PySideConnectionPoller(self)


class IOLoop(QObject):
    def __init__(self, poller):
        self.poller = poller
        self.timers = {}

    def timerEvent(self, event):
        self.timers[event.timerId()](self, timeout_id)

    def add_timeout(self, deadline, callback, oneshot=False):
        deadline = deadline - time.time()
        return PySideTimer(self, callback, oneshot).register(self, deadline)

    def add_soft_timeout(self, min_time, max_time, callback, oneshot=False):
        raise NotImplementedError

    def remove_timeout(self, handler):
        self.timers[timeout_id].unregister(self, handler)

    def stop(self):
        [timer.unregister(self, key) for key, timer in self.timers]
        QTimer.singleShot(0, self.poller.unpoll)
        self.exec_ and QCoreApplication.instance().quit()

    def start(self, exec_=True):
        self.exec_ = exec_
        QTimer.singleShot(0, self.poller.poll)
        self.exec_ and QCoreApplication.instance().exec_()

    def remove_handler(self, fdn=None):
        [notifier.setEnabled(False) for notifier in self.poller]

    def update_handler(self, fdn, event_state):
        self.remove_handler()
        # update notifiers state
        if event_state & READ:
            self.poller.reader.setEnabled(True)
        if event_state & WRITE:
            self.poller.writer.setEnabled(True)
